#include "IOHandler.h"
#include "StringTokenizer.h"

#include <algorithm>     // need transform

// private - remove white space 
void trim(std::string& s)  {
  if ( s.empty() )
    return;
  std::string::iterator p;

  for(p=s.end(),p--; p!= s.begin() && (*p == ' ' || *p == '\t');p--)
    ;
  if ( *p != ' ' && *p != '\t' ) 
    p++;
  s.erase(p,s.end());

  for(p=s.begin(); p!= s.end() && (*p == ' ' || *p == '\t');p++)
    ;
  s.erase(s.begin(),p);

}

// make Indiv struct and initialize it
Indiv *initIndiv(void)  {
  Indiv *ip = new Indiv;
  ip->name = "i";
  ip->pheno = -1;
  ip->covariates = NULL;
  ip->missing = NULL;
  ip->evecs = NULL;
  return ip;
}

// read run parameters and store in map
int IOHandler::readParamFile(char *filename)  {

  ifstream parin;
  parin.open(filename,ios::in);
  if ( !parin.is_open() )  {
    cerr << "can't open file " << filename;
    return 1;
  }

  while ( !parin.eof() )  {
    string line, origline, pname, pval;

    getline(parin,line);
    origline = line;

    if ( line.empty() && parin.eof() )  
      break;

    StringTokenizer st(line,":");
    if ( (int) st.countTokens() != 2 )  {
      cerr << "Can't parse parameter line " << origline;
      return 1;
    }

    st.nextToken(pname);
    st.nextToken(pval);

    // remove leading and trailing white space 
    trim(pname);
    trim(pval);

    if ( !pname.empty() && !pval.empty() )  {
      if ( !pname.compare("SNP") )  {
        snplist.push_back(pval);
      }
      else  {
        setParam(pname,pval);
      }
    }
    else  {
      cerr << "Can't parse parameter line " << origline;
      return 1;
    }

  }
  parin.close();
  return 0;

}


// Get parameter, overwriting if already defined - return 1 if not found
int IOHandler::getParam(string key, string *valp)  {
  map<string,string>::iterator p;
  p = rpar.find(key);
  if ( p != rpar.end() )  {
    *valp = (*p).second;
    return 0;
  }
  else  {
    return 1;
  }
}

// Set parameter, overwriting if already defined
void IOHandler::setParam(string key, string val)  {
  map<string,string>::iterator p;
  p = rpar.find(key);
  if ( p != rpar.end() )  {  // change
    (*p).second = val;
  }
  else  {
    rpar[key] = val;
  }
}


int IOHandler::readIndFile(void)  {

  // parameter "ncovars" has number of covariates (default is zero)
  string numcovars;
  int nc = 0;
  if ( !getParam("ncovars", &numcovars) )  {
    nc = atoi(numcovars.c_str());
  }

  // open indiv file
  string filename;
  getParam("indivfile", &filename);
  ifstream parin;
  parin.open(filename.c_str(),ios::in);
  if ( !parin.is_open() )  {
    cerr << "error: can't open indivfile " << filename << endl;
    return 1;
  }

  int nsamples = 0;
  while ( !parin.eof() )  {
    string line, origline, name, value;

    getline(parin,line);
    origline = line;               // tokenizer will mangle this but error handler may need it

    if ( line.empty() && parin.eof() )  
      break;

    StringTokenizer st(line," \t");

    /*** Put this back in when debugging is done ***
    if ( (int) st.countTokens() != 3 + nc )  {
      cerr << "error: wrong number of tokens on line " << origline << endl;
      return 1;
    }
    */

    Indiv *ip = initIndiv();
    iData.push_back(ip);

    st.nextToken(name);             // get name 
    trim(name);
    ip->name = name;                
    nsamples++;                     

    st.nextToken(value);            // read past sex 
    st.nextToken(value);            // get phenotype
    trim(value);
    
    string str(value);
    transform(str.begin(), str.end(), str.begin(), (int(*)(int)) tolower);

    // phenotype is neither "Case" nor "Control" - need to store name in array anyway, so the 
    // genotypes align with the samples, but pheno == -1 is the signal to ignore it

    if ( !str.compare("case") )  {  // convert phenotype to integer value
      ip->pheno = 1;
    }
    else if ( !str.compare("control") )  {
      ip->pheno = 0;
    }
    else  {
      ip->pheno = -1;
      continue; 
    }

    if ( nc == 0 )  {
      continue;
    }

    // read covariates
    ip->covariates = new double[nc];
    ip->missing = (nc > 0 ? new bool[nc] : NULL);

    for(int i=0;i<nc;i++)  {  
      st.nextToken(value);            
      trim(value);
      if ( !value.compare("NA") )  {
        ip->missing[i] = true;
      }
      else  {
        ip->covariates[i] = atof(value.c_str());
        ip->missing[i] = false;
      }
    }
  }

  parin.close();
  return 0;

}

int IOHandler::readEvecFile(void)  {

  // open evec file
  string filename;
  if ( getParam("evecfile", &filename) )  {
    return 0;  // nothing to read
  }

  string neigstr;
  if ( !getParam("numeigs", &neigstr) )  {
    neigs = atoi(neigstr.c_str());
    if ( neigs == 0 )  
      return 0;  // nothing to read
  }

  ifstream parin;
  parin.open(filename.c_str(),ios::in);
  if ( !parin.is_open() )  {
    cerr << "error: can't open evecfile " << filename << endl;
    return 1;
  }

  int evalline = 1;

  while ( !parin.eof() )  {
    string line, origline, name, value;

    getline(parin,line);
    origline = line;               // tokenizer will mangle this but error handler may need it

    if ( line.empty() && parin.eof() )  
      break;

    if ( evalline )  {            // skip first line (eigenvalues)
      evalline = 0;
      continue;
    }

    StringTokenizer st(line," \t");
    if ( neigs == 0 )  {
      neigs = (int) st.countTokens()-2;
    }
    else if ( (int)st.countTokens()-2 < neigs )  {
      cerr << "error: too few tokens on evec file line " << endl << origline << endl;
      return 1;
    }

    st.nextToken(name);             // get name 
    trim(name);

    Indiv *ip = NULL;
    vector<Indiv *>::iterator p;
    for(p=iData.begin();p != iData.end();p++)  {
      Indiv *jp = *p;
      if ( !jp->name.compare(name)  )  {
        ip = jp;
        break;
      }
    } 

    if ( ip == NULL )  {
      cerr << "warning: evec file individual " << name << "is not in the data set " << endl;
      continue;
    }
    ip->evecs = new double[neigs];

    for(int i=0;i<neigs; i++)  {
      st.nextToken(value);
      trim(value);
      ip->evecs[i] = atof(value.c_str());
    }

  }
  parin.close();

  return 0;

}


int IOHandler::getNcovars(void)  {
  string numcovars;
  int nc = 0;
  if ( !getParam("ncovars", &numcovars) )  {
    nc = atoi(numcovars.c_str());
  }
  return nc;
}

int IOHandler::getNeigs(void)  {
  return neigs;
}

int IOHandler::readCovariateParameters(void)  {
  int n = getNcovars() + snplist.size();

  if ( n == 0 )  {
    return 0;
  }

  string filename;
  if ( getParam("covarfile",&filename) )  {
    cerr << "error: no covarfile in parameter file" << endl;
    return 1;
  }

  beta = new double[n];
  tbar = new double[n];

  ifstream fin;
  fin.open(filename.c_str(),ios::in);
  if ( !fin.is_open() )  {
    cerr << "Can't open input file " << filename << endl;
    return 1;
  }

  // tentative input format 
  // <first-line> m-value
  // <next ncovar lines> tbar[i], beta[i]

  int count = -1;
  while ( !fin.eof() )  {

    string line, origline, token;
    getline(fin,line);

    StringTokenizer stt(line," \t");

    if ( line.empty() || fin.eof() )
      break;

    if ( count >= n )  {
      cerr << "error: wrong number of lines in " << filename << endl;
      return 1;
    }

    if ( count == -1 )  {
      stt.nextToken(token);
      mval = (double) atof(token.c_str());
    }
    else  {
      stt.nextToken(token);
      tbar[count] = (double) atof(token.c_str());
      stt.nextToken(token);
      beta[count] = (double) atof(token.c_str());
    }
    count++;

  }
  if ( count != n )  {
    cout << "error: covariate parameter file has wrong number of lines" << endl;
    return 1;
  }

  return 0;

}

int *IOHandler::findSNPs(const char *filename)  {

  ifstream parin;
  parin.open(filename,ios::in);
  if ( !parin.is_open() )  {
    cerr << "can't open file " << filename << endl;
    return NULL;
  }

  map<string,int> indices;
  vector<string>::iterator vp;
  for(vp=snplist.begin();vp!=snplist.end();vp++)  {
    indices[*vp] = -1;    // initialize to "not found"
  }

  int linenum = -1;
  map<string,int>::iterator mp;
  while ( !parin.eof() )  {
    string line, origline, pname, pval;

    linenum++;
    getline(parin,line);
    origline = line;

    if ( line.empty() && parin.eof() )
      break;

    StringTokenizer st(line," \t");
    if ( (int) st.countTokens() < 2 )  {
      cerr << "Can't parse parameter line " << origline;
      return NULL;
    }

    st.nextToken(pname);
    trim(pname);

    mp = indices.find(pname);
    if ( mp != indices.end() )  {
      (*mp).second = linenum;
    }

  }
  parin.close();

  // pack indices into an array
  int *index = new int[snplist.size()];
  int n = 0;
  for(vp=snplist.begin();vp!=snplist.end();vp++)  {
    mp = indices.find(*vp);
    if ( (*mp).second == -1 )  {
      cerr << "warning: " << (*vp) << " is not in the SNP file, will be ignored" << endl;
    }
    index[n++] = (*mp).second;  // note: -1 in array signifies "not found"
  }
  return index;

}



void IOHandler::dbgShowParams(void)  {
  map<string, string>::iterator p;
  for(p=rpar.begin();p != rpar.end(); p++)  {
    cout << "PARAMETER:" << p->first << "=" << p->second << endl;
  }
}


void IOHandler::dbgShowIndivs(void)  {

  string numcovars;
  int nc;
  if ( !getParam("ncovars", &numcovars) )  {
    nc = atoi(numcovars.c_str());
  }

  vector<Indiv *>::iterator p;
  for(p=iData.begin();p != iData.end();p++)  {
    Indiv *ip = *p;
    cout << "INDIV:" << ip->name; 
    if ( ip->pheno == 0 )  {
      cout << " Case ";
    }
    else if ( ip->pheno == 1 )  {
      cout << " Control ";
    }
    else  {
      cout << "(ignore)" << endl;
      continue;
    }

    cout << "C = [ ";
    for(int i=0;i<nc;i++)  { 
      if ( ip->missing[i] )  {
        cout << "-";
      }
      else  {
        cout << setw(10) << setprecision(4) << ip->covariates[i];
      }
      cout << "  ";
    }
    cout << " ]" << endl;
  }
}

void IOHandler::dbgShowCovarParams(void)  {
  int n = getNcovars();
  cout << "COVARIATE PARAMETERS  " << endl;
  if ( n == 0 )  {
    cout << "(none)" << endl;
  }

  cout << "m = " << setw(10) << setprecision(4) << mval;
  for(int i=0;i<n;i++)  {
    cout << "  PARAM " << i << " (beta,tbar) = { " 
         << beta[i] << " , " << tbar[i] << "} ";
  }
  cout << endl;
}





