#include "IOHandler2.h"
#include "StringTokenizer.h"

#include <algorithm>     // need transform

// private - remove white space 
void trim(std::string& s)  {
  if ( s.empty() )
    return;
  std::string::iterator p;

  for(p=s.end(),p--; p!= s.begin() && (*p == ' ' || *p == '\t');p--)
    ;
  if ( *p != ' ' && *p != '\t' ) 
    p++;
  s.erase(p,s.end());

  for(p=s.begin(); p!= s.end() && (*p == ' ' || *p == '\t');p++)
    ;
  s.erase(s.begin(),p);

}

// read run parameters and store in map
int IOHandler::readParamFile(char *filename)  {

  ifstream parin;
  parin.open(filename,ios::in);
  if ( !parin.is_open() )  {
    cerr << "can't open file " << filename;
    return 1;
  }

  while ( !parin.eof() )  {
    string line, origline, pname, pval;

    getline(parin,line);
    origline = line;

    if ( line.empty() && parin.eof() )  
      break;

    StringTokenizer st(line,":");
    if ( (int) st.countTokens() != 2 )  {
      cerr << "Can't parse parameter line " << origline;
      return 1;
    }

    st.nextToken(pname);
    st.nextToken(pval);

    // remove leading and trailing white space 
    trim(pname);
    trim(pval);
    
    if ( !pname.empty() && !pval.empty() )  {
      setParam(pname,pval);
    }
    else  {
      cerr << "Can't parse parameter line " << origline;
      return 1;
    }

  }
  parin.close();
  return 0;

}


// Get parameter, overwriting if already defined - return 1 if not found
int IOHandler::getParam(string key, string *valp)  {
  map<string,string>::iterator p;
  p = rpar.find(key);
  if ( p != rpar.end() )  {
    *valp = (*p).second;
    return 0;
  }
  else  {
    return 1;
  }
}

// Set parameter, overwriting if already defined
void IOHandler::setParam(string key, string val)  {
  map<string,string>::iterator p;
  p = rpar.find(key);
  if ( p != rpar.end() )  {  // change
    (*p).second = val;
  }
  else  {
    rpar[key] = val;
  }
}


int IOHandler::readIndFile(void)  {

  // parameter "ncovars" has number of covariates (default is zero)
  string numcovars;
  int nc;

  if ( !getParam("ncovars", &numcovars) )  {
    nc = atoi(numcovars.c_str());
  }
  else  {
    nc = 0;
  }

  // open indiv file
  string filename;
  getParam("indivfile", &filename);
  ifstream parin;
  parin.open(filename.c_str(),ios::in);
  if ( !parin.is_open() )  {
    cerr << "error: can't open indivfile " << filename << endl;
    return 1;
  }

  int nsamples = 0;
  while ( !parin.eof() )  {
    string line, origline, name, value;

    getline(parin,line);
    origline = line;               // tokenizer will mangle this but error handler may need it

    if ( line.empty() && parin.eof() )  
      break;

    StringTokenizer st(line," \t");
    if ( (int) st.countTokens() < 3 + nc )  {
      cerr << "error: wrong number of tokens on line " << origline << endl;
      return 1;
    }

    Indiv *ip = new Indiv; 
    iData.push_back(ip);

    st.nextToken(name);             // get name 
    trim(name);

    ip->name = name;                
    nsamples++;                     

    st.nextToken(value);            // read past sex 
    st.nextToken(value);            // get phenotype
    trim(value);
    
    string str(value);
    transform(str.begin(), str.end(), str.begin(), (int(*)(int)) tolower);

    // phenotype is neither "Case" nor "Control" - need to store name in array anyway, so the 
    // genotypes align with the samples, but pheno == -1 is the signal to ignore it

    if ( !str.compare("case") )  {  // convert phenotype to integer value
      ip->pheno = 1;
    }
    else if ( !str.compare("control") )  {
      ip->pheno = 0;
    }
    else  {
      ip->pheno = -1;
      continue; 
    }

    // read covariates
    ip->covariates = new double[nc];
    ip->missing = new bool[nc];
    for(int i=0;i<nc;i++)  {  
      st.nextToken(value);            
      trim(value);
      if ( !value.compare("NA") )  {
        ip->missing[i] = true;
      }
      else  {
        ip->covariates[i] = atof(value.c_str());
        ip->missing[i] = false;
      }
    }

  }
  parin.close();
  return 0;

}


int IOHandler::getNcovars(void)  {
  string numcovars;
  int nc = -1;
  if ( !getParam("ncovars", &numcovars) )  {
    nc = atoi(numcovars.c_str());
  }
  return nc;
}

int IOHandler::readCovariateParameters(void)  {

  string filename;
  if ( getParam("covarfile",&filename) )  {
    cerr << "error: no covarfile in parameter file" << endl;
    return 0;
  }

  int n = getNcovars();

  beta = new double[n];
  tbar = new double[n];

  ifstream fin;
  fin.open(filename.c_str(),ios::in);
  if ( !fin.is_open() )  {
    cerr << "Can't open input file " << filename << endl;
    return 1;
  }

  // tentative input format 
  // <first-line> m-value
  // <next ncovar lines> tbar[i], beta[i]

  int count = -1;
  while ( !fin.eof() )  {

    string line, origline, token;
    getline(fin,line);

    StringTokenizer stt(line," \t");

    if ( line.empty() || fin.eof() )
      break;

    if ( count >= n )  {
      cerr << "error: wrong number of lines in " << filename << endl;
      return 1;
    }

    if ( count == -1 )  {
      stt.nextToken(token);
      mval = (double) atof(token.c_str());
    }
    else  {
      stt.nextToken(token);
      tbar[count] = (double) atof(token.c_str());
      stt.nextToken(token);
      beta[count] = (double) atof(token.c_str());
    }
    count++;

  }

  return 0;

}


void IOHandler::dbgShowParams(void)  {
  map<string, string>::iterator p;
  for(p=rpar.begin();p != rpar.end(); p++)  {
    cout << "PARAMETER:" << p->first << "=" << p->second << endl;
  }
}


void IOHandler::dbgShowIndivs(void)  {

  string numcovars;
  int nc;
  if ( !getParam("ncovars", &numcovars) )  {
    nc = atoi(numcovars.c_str());
  }

  vector<Indiv *>::iterator p;
  for(p=iData.begin();p != iData.end();p++)  {
    Indiv *ip = *p;
    cout << "INDIV:" << ip->name; 
    if ( ip->pheno == 0 )  {
      cout << " Case ";
    }
    else if ( ip->pheno == 1 )  {
      cout << " Control ";
    }
    else  {
      cout << "(ignore)" << endl;
      continue;
    }

    cout << "C = [ ";
    for(int i=0;i<nc;i++)  { 
      if ( ip->missing[i] )  {
        cout << "-";
      }
      else  {
        cout << setw(10) << setprecision(4) << ip->covariates[i];
      }
      cout << "  ";
    }
    cout << " ]" << endl;
  }
}

void IOHandler::dbgShowCovarParams(void)  {
  int n = getNcovars();
  cout << "COVARIATE PARAMETERS  " << endl;
  cout << "m = " << setw(10) << setprecision(4) << mval;
  for(int i=0;i<n;i++)  {
    cout << "  PARAM " << i << " (beta,tbar) = { " 
         << beta[i] << " , " << tbar[i] << "} ";
  }
  cout << endl;
}





