#include "LTFitIOHandler.h"
#include "LTFit.h"
#include "LinearSolve.h"

#define ROOT2 1.41421356237310
#define ROOT2PI 2.506628274631

// -------------------------------------------------------------------------------------------
// Clinical covariates
// -------------------------------------------------------------------------------------------
// From tbar, t-values and prevalence (stored in sigmas on input) find 
// effect sizes and threshold.
// -------------------------------------------------------------------------------------------

int LTFit::fitClinicalCovariateData(void)  {

  covariates = ioh->getCovariates();
  ncovar = ioh->getNcovars();

  Covariate *cp;
  DataPoint *dp;
  double y;

  // compute sigmas from prevalence
  for(int i=0;i<ncovar;i++)  {
    cp = covariates[i];
    for(int k=0;k<cp->ndatapts;k++)  {
      dp = cp->data[k];
      if ( inverseDnorm(dp->sigma, &y) )  {
        cerr << "error: can't computing liability" << endl;
        return 1;
      }
      dp->sigma = y;
    }
  }

  // dbgShowInput();

  leastSquaresSolution();
  return 0;

}

void LTFit::leastSquaresSolution(void)  {

  int L = 1;
  for(int i=0;i<ncovar;i++)  {
    L += covariates[i]->nindices;
  }

  double *A = new double[L*L];
  double *b = new double[L];
  double *v = new double[L];

  // initialize
  for(int i=0;i<L;i++)  {
    b[i] = 0.0;
    v[i] = 0.0;
    for(int j=0;j<L;j++)  {
      A[i*L+j] = 0.0;
    }
  }

  // upper-left elements
  int l = 0;
  for(int i1=0;i1<ncovar;i1++)  {
    for(int k1=0;k1<covariates[i1]->nindices;k1++)  {

      int m = 0;
      for(int i2=0;i2<ncovar;i2++)  {
        for(int k2=0;k2<covariates[i2]->nindices;k2++)  {
          if ( i1 == i2 )   {

            double a = 0.0;
            for(int j=0;j<covariates[i1]->ndatapts;j++)  {
              double z = (covariates[i1]->data[j]->T[k1] - covariates[i1]->tbar[k1]);
              double y = (covariates[i2]->data[j]->T[k2] - covariates[i2]->tbar[k2]);
              a += y*z;
            }

            A[l*L+m] = a;
          }
          m++;

        }
      }

      l++;
    }
  }

  // first L-1 rows, last column (coefficient of m in equation for C_ik) and
  // last row, first L-1 columns, last column (coefficients of C_ik in equation for m) and
  // first L-1 rows of right-side vector (constant in equation or C_ik)

  l = 0;
  for(int i=0;i<ncovar;i++)  {
    for(int k=0;k<covariates[i]->nindices;k++)  {
      double a = 0.0;
      double e = 0.0;
      for(int j=0;j<covariates[i]->ndatapts;j++)  {
        double z = (covariates[i]->data[j]->T[k] - covariates[i]->tbar[k]);
        a += z;
        e += z*covariates[i]->data[j]->sigma;
      }
      A[l*L+(L-1)] = a;     // note: some of the signs have been moved around
      A[(L-1)*L+l] = a;     // to make LTSOFT-0.6 reproduce LTSOFT-0.5 results
      b[l] = -e;
      l++;
    }
  }

  // lower-left element in matrix (coefficient of m in equation for m)
  int ia = 0;
  for(int i=0;i<ncovar;i++)  {
    ia += covariates[i]->ndatapts;
  }
  A[(L-1)*L+(L-1)] = (double) ia;

  // last element in right-side vector
  double a = 0.0;
  for(int i=0;i<ncovar;i++)  {
    for(int j=0;j<covariates[i]->ndatapts;j++)  {
      a += covariates[i]->data[j]->sigma;
    }
  }
  b[L-1] = -a;

  LinearSolve(A,b,v,L);

  l = 0;
  for(int i=0;i<ncovar;i++)  {
    for(int k=0;k<covariates[i]->nindices;k++)  {
      a += covariates[i]->C[k] = -v[l];
      l++;
    }
  }
  threshold = v[L-1];

  free(A);
  free(b);
  free(v);

}

// binary search for inverse of R function dnorm - convert prevalence to sigma (quantile)
#define BIG 100.0
#define TINY 1.0e-10
#define MAXITER 1000
int LTFit::inverseDnorm(double y, double *result)  {
  double xl  = -BIG;
  double xr = +BIG;
  double fl = dnorm(xl) - y;
  double fr = dnorm(xr) - y;
  int iter = 0;

  while ( fl*fr > 0.0 )  {      // if HUGE isn't large enough, expand the search interval
    xl *= 2.0;
    xr *= 2.0;
    fl = dnorm(xl) - y;
    fr = dnorm(xr) - y;
    if ( iter++ > MAXITER )  {
      cerr << "error in binary search - can't find search interval"
           << "y = " << y << endl;
      return 1;
    }
  }

  double xmid = 0.5*(xl+xr);
  double fmid = dnorm(xmid) - y;
  iter = 0;
  while ( fabs(fmid) > TINY )  {
    if ( fl*fmid < 0 )  {
      xr = xmid;
      fr = fmid;
    }
    else if ( fmid*fr < 0 )  {
      xl = xmid;
      fl = fmid;
    }
    xmid = 0.5*(xl+xr);
    fmid = dnorm(xmid) - y;
    if ( iter++ > MAXITER )  {
      cerr << "error in binary search - can't find root" << endl;
      return 1;
    }
  }
  *result = xmid;
  return 0;

}

// R function dnorm = integral from -Inf of Gaussian(0,1)
double LTFit::dnorm(double x)  {
  double y;
  if ( x > 0 )  {
    y = 1.0 + erf(x/ROOT2);
  }
  else  {
    y = 1.0 - erf(-x/ROOT2);
  } 
  y *= 0.5;
  return(y);
}

// ----------------------------------------------------------------------------------
//  (dead code?)
//  Theory: This is the functional we are minimizing in the least-squares fit
// ----------------------------------------------------------------------------------
double LTFit::evalF(void)  {

  double f = 0.0;
  for(int i=0;i<ncovar;i++)  {
    for(int j=0;j<covariates[i]->ndatapts;j++)  {

      double sigma_ij = covariates[i]->data[j]->sigma;
      double z = sigma_ij;
      for(int k=0;k<covariates[i]->nindices;k++)  {

        double tbar_ik = covariates[i]->tbar[k];
        double c_ik = covariates[i]->C[k];
        double t_ijk = covariates[i]->data[j]->T[k];
        z -= c_ik*(t_ijk - tbar_ik);

      }
      z += threshold;
      f  += (z*z);

    }
  }
  return f;

}

// ----------------------------------------------------------------------------------
//  Theory: This is the functional we are minimizing in the least-squares fit,
//  revised for the possibility that we have covariate pairs.
// ----------------------------------------------------------------------------------
double LTFit::evalF(double *cval)  {
  
  double f = 0.0;
  
  // put array values into structs
  int l = 0;
  for(int i=0;i<ncovar;i++)  {
    for(int k=0;k<covariates[i]->nindices;k++)  {
      covariates[i]->C[k] = cval[l++];
    } 
  }        
  threshold = cval[l];
    
  for(int i=0;i<ncovar;i++)  {
    for(int j=0;j<covariates[i]->ndatapts;j++)  {
  
      double sigma_ij = covariates[i]->data[j]->sigma;
      double z = sigma_ij;
      for(int k=0;k<covariates[i]->nindices;k++)  {
    
        double tbar_ik = covariates[i]->tbar[k];
        double c_ik = covariates[i]->C[k];
        double t_ijk = covariates[i]->data[j]->T[k];
        z -= c_ik*(t_ijk - tbar_ik);
      
      }
      z += threshold;
      f  += (z*z);
    
    }
  }   
      
  return f;
  
} 



// Write out clinical covariate part of the covariate parameter file (input to LTScore)
// format is (1) first line <threshold> (2) one line per covariate in form
// <mean0> <effect-size0> then possibly <mean1> <effect-size1>

void LTFit::writeClinicalCovariateData(const char *filename)  {

  if ( !outisopen )  {
    outisopen = true;
    outstream.open(filename,ios::out);
  }

  outstream << fixed << setw(10) << setprecision(4) << -threshold << endl;
  thoutput = true;
  for(int i=0;i<ncovar;i++)  {
    for(int k=0;k<covariates[i]->nindices;k++)  {
      outstream << fixed << setw(10) << setprecision(6) << covariates[i]->tbar[k] << "  ";
      outstream << fixed << setw(10) << setprecision(6) << covariates[i]->C[k] << "  ";
    }
    outstream << endl;
  }

}

// -------------------------------------------------------------------------------------------
// Genetic covariates
// -------------------------------------------------------------------------------------------

int LTFit::fitGeneticCovariateData()  {

  phenotypes = ioh->getPhenotypes();
  nindiv = ioh->getNindiv();
  vector<string> snplist = ioh->getSnplist();
  int nsnps = snplist.size();

  // index is an array of the line numbers in the .geno file at which
  // the genetic covariates are located, not necessarily in order

  int *index;
  string f;
  if ( !ioh->getParam("snpfile",&f) )  {
    index = ioh->findSNPs(f.c_str());
    if ( index == NULL )  {
      return 1;
    }
  }
  else  {
    cerr << "error: SNP file must be supplied for genetic covariates" << endl;
    return 1;
  }

  // sort indices (assume few so use bubble-sort)
  for(int i=1;i<nsnps;i++)  {
    for(int j=0;j<i;j++)  {
      if ( index[j] > index[i] )  {
        int t = index[j];
        index[j] = index[i];
        index[i] = t;
      }
    }
  }
  // squeeze out "not found"
  int i=0;
  while ( index[i] == -1 )  {
    for(int j=1;j<nsnps;j++)  {
      index[j-1] = index[j];
    }
    nsnps--;
  }

  // open genotype file for input
  string gfilename;
  if ( ioh->getParam("genotypefile",&gfilename) )  {
    cerr << "error: no genotypefile in parameter file" << endl;
    return 1;
  }

  ifstream genoinstream;
  genoinstream.open(gfilename.c_str(),ios::in);
  if ( !genoinstream.is_open() )  {
    cerr << "error: can't open genotypefile " << gfilename << endl;
    return 1;
  }

  // open output file
  if ( !outisopen )  {
    string ofilename;
    if ( ioh->getParam("outfile",&ofilename) )  {
      cerr << "error: no output file in parameter file" << endl;
      return 1;
    }

    outstream.open(ofilename.c_str(),ios::out);
    if ( !outstream.is_open() )  {
      cerr << "error: can't open score output file " << ofilename << endl;
      return 1;
    }
    outisopen = true;
  }

  // if user specifies a prevalence, use it
  string s;
  if ( !ioh->getParam("prevalence",&s) )  {
    prevalence = atof(s.c_str());
    if ( inverse(prevalence, &threshold) )  {
      cerr << "error: can't find inverse of prevalence" << endl;
      return 1;
    }
    if ( !thoutput )  {
      outstream << fixed << setw(10) << setprecision(4) << threshold << endl;
    }
  }
  else  {
    if (!thresholdisset) {
      cerr << "error: parameter file does not have prevalence " 
           << "and no clinical covariates to deduce from" << endl;
      return 1;
    }
    else  {
      prevalence = exp(-0.5*threshold*threshold)/ROOT2PI;
    }
  }

//  if ( !thresholdisset )  {
//    string s;
//    if ( ioh->getParam("prevalence",&s) )  {
//      cerr << "error: parameter file does not list prevalence" << endl;
//      return 1;
//    }
//    prevalence = atof(s.c_str());
//    if ( inverse(prevalence, &threshold) )  {
//      cerr << "error: can't find inverse of prevalence" << endl;
//      return 1;
//    }
//    outstream << fixed << setw(10) << setprecision(6) << threshold << endl;
//  }
//  else  {
//    prevalence = exp(-0.5*threshold*threshold)/ROOT2PI;
//  }

  threshold = fabs(threshold);
  // cout << "DEBUG (prevalence, threshold) " << prevalence << "  " << threshold << endl;

  int *G = new int[nindiv];
  int snpnum = -1;
  int snpidx = 0;

  while ( !genoinstream.eof() )  {
    snpnum++;

    if ( snpidx >= nsnps ) {                      // no more genetic covariates to process
      break;
    }

    if ( readGenotype(G, nindiv, &genoinstream) ) continue;

    if ( snpnum < index[snpidx] )  {     // this SNP is not a genetic covariate
      continue;
    }

    double sumcase = 0.0, sumctrl = 0.0;
    int countcase = 0, countctrl = 0;
    for(int i=0;i<nindiv;i++)  {
      if ( G[i] == 9 )
        continue;
      if ( phenotypes[i] == 1 )  {       // case
        sumcase += (double) G[i];
        countcase += 2;
      }
      else if ( phenotypes[i] == 0 )  {  // ctrl
        sumctrl += (double) G[i];
        countctrl += 2;
      }
    }

    fcase = (countcase > 0 ? sumcase/countcase : 0.0);
    fctrl = (countctrl > 0 ? sumctrl/countctrl : 0.0);
    fexp = prevalence*fcase + (1.0-prevalence)*fctrl;

    int istat = 0;
    double epsilon = minimize(&istat);
    outstream << fixed << setw(10) << setprecision(6)
              << fexp << "  " << epsilon << endl;


    snpidx++;

  }

  outstream.close();
  delete G;
  delete index;
  return 0;

}


int LTFit::readGenotype(int *g, int n, ifstream *gin)  {
  if ( !gin->is_open() )  {
    cerr << "error: genotype file not open" << endl;
    return 1;
  }

  string line;
  getline(*gin,line);

  if ( gin->eof() ) {
    return -1;
  }

  if ( (int) line.length() < n)  {
    cerr << "error: line too short" << endl;
    return 1;
  }

  if ( (int) line.length() > n)  {
    cerr << "(warning) line too long length = " << line.length() << " n = " << n << endl;
  }

  for(int i=0;i<n;i++)  {
    int k = (int) line[i];
    k -= 48;                     // 48 is ASCII('0')
    if ( 0 <= k && k <= 2 )  {
      g[i] = k;
    }
    else  {
      g[i] = 9;
    }
  }
  return 0;
}

double LTFit::minimize(int *ierr)  {

  // Telescoping Grid search
  int N = 5;

  // set initial search interval
  double xl = -1.0, xr = 1.0, x0 = 0.0, dx = (xr - xl)/N;

  // find three points (xl,x0,xr) such that f(xl) > f(x0) and f(xr) > f(x0)

  int dbg = 0, done = 0;
  while ( !done && ++dbg < 200 )  {

    double fl = f(xl);
    double f0 = f(x0);
    double fr = f(xr);

    if ( fl < f0 && f0 > fr )  {      //  local maximum?
      exit(1);
    }
    else if ( fl >= f0 && f0 >= fr )  {    // decreasing
      xl = x0;
      x0 = xr;
      xr += dx;
    }
    else if ( fl <= f0 && f0 <= fr )  {    // increasing
      xr = x0;
      x0 = xl;
      xl -= dx;
    }
    else  {
      done = 1;
    }
  }
  if ( !done )  {
    cout << "error: can't find an interval with a minimum on it" << endl;
    return 0.0;
  }

  do  {

    double fr = f(xr);
    double fl = f(xl);
    double fmin = ( fr < fl ? fr : fl);
    int imin = 0;

    for(int i=1;i<=N;i++)  {

      double x = xl + i*dx;
      double fx = f(x);

      if ( fx < fmin )  {
        imin = i;
        fmin = fx;
      }
    }

    if ( imin == 0 )  {
      xr = xl+dx;
    }
    else if ( imin == N )  {
      xl = xr-dx;
    }
    else  {
      xl = xl + (imin-1)*dx;
      xr = xl + (imin+1)*dx;
    }
    dx = (xr - xl)/N;



  } while ( dx > 1e-6 );

  return xl+0.5*dx;

}

// This is R function qnorm(x,0,1)
#define BIG 100.0
#define TINY 1.0e-10 
#define MAXITER 1000
int LTFit::inverse(double y, double *result)  {
  double xl  = -BIG;
  double xr = +BIG;
  double fl = dnorm(xl) - y;
  double fr = dnorm(xr) - y;
  int iter = 0; 
        
  while ( fl*fr > 0.0 )  {      // if HUGE isn't large enough, expand the search interval
    xl *= 2.0;
    xr *= 2.0;
    fl = dnorm(xl) - y;
    fr = dnorm(xr) - y;
    if ( iter++ > MAXITER )  {
      cerr << "error in binary search - can't find search interval"
           << "y = " << y << endl;
      return 1;
    }
  } 
      
  double xmid = 0.5*(xl+xr);
  double fmid = dnorm(xmid) - y;
  iter = 0;
  while ( fabs(fmid) > TINY )  {
    if ( fl*fmid < 0 )  {
      xr = xmid;
      fr = fmid;
    }
    else if ( fmid*fr < 0 )  {
      xl = xmid;
      fl = fmid;
    }
    xmid = 0.5*(xl+xr);
    fmid = dnorm(xmid) - y;
    if ( iter++ > MAXITER )  {
      cerr << "error in binary search - can't find root" << endl;
      return 1;
    }
  }
  *result = xmid;
  return 0;

}

// R function pnorm
double LTFit::pnorm(double x, double mu, double sigma, int lower_tail)  {
  double z = (x - mu)/sigma;
  double y = 0.5*(1.0 + erf(z/ROOT2));

  if ( lower_tail )  {
    return y;
  }
  else  {
    return 1.0-y;
  }
}

double LTFit::f(double x)  {

  double p = fexp/2;
  double q = 1.0-p;
  double psd = sqrt(2*p*(1-p));
  double sigma, mu, pctrl1, pctrl2, pcase1, pcase2, result, expfca, expfco;

  sigma = sqrt(1-x*x);
  mu = x*(1-2*p)/psd;
  pctrl1 = pnorm(threshold, mu, sigma, 1);

  mu = x*(2-2*p)/psd;
  pctrl2 = pnorm(threshold, mu, sigma, 1);

  pcase1 = 1.0 - pctrl1;
  pcase2 = 1.0 - pctrl2;

  pctrl1 *= ( 2*p*q/(1.0-prevalence) );
  pctrl2 *= ( p*p/(1.0-prevalence) );

  pcase1 *= ( 2*p*q/prevalence );
  pcase2 *= ( p*p/prevalence );

  expfca = (pcase1 + 2*pcase2);
  expfco = (pctrl1 + 2*pctrl2);

  result = (pctrl1 + 2*pctrl2 - fctrl)*(pctrl1 + 2*pctrl2 - fctrl) +
           (pcase1 + 2*pcase2 - fcase)*(pcase1 + 2*pcase2 - fcase);

  return (result);

}

