#include "LTFitIOHandler.h"
#include "StringTokenizer.h"
#include "Covariate.h"

#include <algorithm>     // need transform

// private - remove white space 
void trim(std::string& s)  {
  if ( s.empty() )
    return;
  std::string::iterator p;

  for(p=s.end(),p--; p!= s.begin() && (*p == ' ' || *p == '\t');p--)
    ;
  if ( *p != ' ' && *p != '\t' ) 
    p++;
  s.erase(p,s.end());

  for(p=s.begin(); p!= s.end() && (*p == ' ' || *p == '\t');p++)
    ;
  s.erase(s.begin(),p);

}

// read run parameters and store in map
int LTFitIOHandler::readParamFile(const char *filename)  {

  ifstream parin;
  parin.open(filename,ios::in);
  if ( !parin.is_open() )  {
    cerr << "can't open file " << filename;
    return 1;
  }

  while ( !parin.eof() )  {
    string line, origline, pname, pval;

    getline(parin,line);
    origline = line;

    if ( line.empty() && parin.eof() )  
      break;

    StringTokenizer st(line,":");
    if ( (int) st.countTokens() != 2 )  {
      cerr << "Can't parse parameter line " << origline;
      return 1;
    }

    st.nextToken(pname);
    st.nextToken(pval);

    // remove leading and trailing white space 
    trim(pname);
    trim(pval);

    if ( !pname.empty() && !pval.empty() )  {
      if ( !pname.compare("SNP") )  {
        snplist.push_back(pval);
      }
      else  {
        setParam(pname,pval);
      }
    }
    else  {
      cerr << "Can't parse parameter line " << origline;
      return 1;
    }

  }
  parin.close();
  return 0;

}

int *LTFitIOHandler::findSNPs(const char *filename)  {

  ifstream parin;
  parin.open(filename,ios::in);
  if ( !parin.is_open() )  {
    cerr << "can't open file " << filename << endl;
    return NULL;
  }

  map<string,int> indices;
  vector<string>::iterator vp;
  for(vp=snplist.begin();vp!=snplist.end();vp++)  {
    indices[*vp] = -1;    // initialize to "not found"
  }

  int linenum = -1;
  map<string,int>::iterator mp;
  while ( !parin.eof() )  {
    string line, origline, pname, pval;

    linenum++;
    getline(parin,line);
    origline = line;

    if ( line.empty() && parin.eof() )  
      break;

    StringTokenizer st(line," \t");
    if ( (int) st.countTokens() < 2 )  {
      cerr << "Can't parse parameter line " << origline;
      return NULL;
    }

    st.nextToken(pname);
    trim(pname);

    mp = indices.find(pname);
    if ( mp != indices.end() )  {
      (*mp).second = linenum;
    }

  }
  parin.close();

  // pack indices into an array
  int *index = new int[snplist.size()];
  int n = 0;
  for(vp=snplist.begin();vp!=snplist.end();vp++)  {
    mp = indices.find(*vp);
    if ( (*mp).second == -1 )  {
      cerr << "warning: " << (*vp) << " is not in the SNP file, will be ignored" << endl;
    }
    index[n++] = (*mp).second;  // note: -1 in array signifies "not found"
  }
  return index;

}




int LTFitIOHandler::readDataFile(const char *filename)  {

  ifstream fin;
  fin.open(filename,ios::in);
  if ( !fin.is_open() )  {
    cerr << "Can't open input file " << filename << endl;
    return 1;
  }

  // temporary vectors until we know how large to allocate arrays
  vector<Covariate *> cvtmp;
  vector<DataPoint *> dptmp;

  ncovar = 0;
  while ( !fin.eof() )  {

    Covariate *cp;
    DataPoint *dp;

    string line, origline, token;
    getline(fin,line);                      // get next line

    if ( line.empty() || fin.eof()  )  {
      if ( !dptmp.empty() )  {               // put data points into previous covariate set
        int ndp = dptmp.size();
        cp->data = new DataPoint *[ndp];
        cp->ndatapts = ndp;
        for(int i=0;i<ndp;i++)  {
           cp->data[i] = dptmp[i];
        }
        dptmp.clear();
      }
      break;
    }

    StringTokenizer stt(line," \t");
    stt.nextToken(token);                   //  get first token on line

    if ( !token.compare("MEAN") )  {        //  begin new covariate section
      if ( !dptmp.empty() )  {               // put data points into previous covariate set
        int ndp = dptmp.size();
        cp->data = new DataPoint *[ndp];
        cp->ndatapts = ndp;
        for(int i=0;i<ndp;i++)  {
           cp->data[i] = dptmp[i];
        }
        dptmp.clear();
      }

      ncovar++;
      int nind = stt.countTokens()-1;
      cp = new Covariate(nind);
      cvtmp.push_back(cp);

      for(int i=0;i<nind;i++)  {            // read means
        stt.nextToken(token);
        cp->tbar[i] = atof(token.c_str());
        cp->C[i] = 0.0;
      }

    }
    else  {                                       // read data point into current covariate section
      if ( stt.countTokens() != (unsigned) (cp->nindices)+1 )  {       // should be nind covariates plus one prevalence
        cerr << "Error in data input" << endl;
        return 1;
      }

      dp = new DataPoint(cp->nindices);
      dptmp.push_back(dp);
      for(int i=0;i<cp->nindices;i++)  {    // read covariates
        if ( i > 0 )                        // first token on line already read
          stt.nextToken(token);
        if ( !token.compare("NA") )  {
          dp->missing[i] = 1;
          dp->T[i] = 0.0;
        }
        else  {
          dp->missing[i] = 0;
          dp->T[i] = atof(token.c_str());
        }
      }
      stt.nextToken(token);          // read prevalence
      dp->sigma = atof(token.c_str());       // store it in sigma field for now
    }

  }

  covariates = new Covariate *[ncovar];
  for(int i=0;i<ncovar;i++)  {
    covariates[i] = cvtmp[i];
  }

  fin.close();
  return 0;

}

int LTFitIOHandler::readIndFile(void)  {

  // open indiv file
  string filename;
  getParam("indivfile", &filename);
  ifstream parin;
  parin.open(filename.c_str(),ios::in);
  if ( !parin.is_open() )  {
    cerr << "error: can't open indivfile " << filename << endl;
    return 1;
  }

  nindiv = 0;
  vector<int> pheno;

  while ( !parin.eof() )  {
    string line, origline, name, value;
    
    getline(parin,line);
    origline = line;               // tokenizer will mangle this but error handler may need it
    
    if ( line.empty() && parin.eof() )
      break;

    StringTokenizer st(line," \t");
    if ( (int) st.countTokens() < 3 )  {
      cerr << "error: wrong number of tokens on line " << origline << endl;
      return 1;
    }
  
    st.nextToken(name);             // read past name
    st.nextToken(value);            // read past sex
    st.nextToken(value);            // get phenotype
    trim(value);
    nindiv++;
    
    string str(value);
    transform(str.begin(), str.end(), str.begin(), (int(*)(int)) tolower);
    
    // phenotype is neither "Case" nor "Control" - need to store name in array anyway, so the
    // genotypes align with the samples, but pheno == -1 is the signal to ignore it
    if ( !str.compare("case") )     // convert phenotype to integer value
      pheno.push_back(1);
    else if ( !str.compare("control") )  
      pheno.push_back(0);
    else  
      pheno.push_back(-1);

  }
  parin.close();

  phenotypes = new int[nindiv];
  int i = 0;
  vector<int>::iterator p;
  for(p=pheno.begin();p!=pheno.end();p++)  {
    phenotypes[i++] = *p;
  }

  return 0;

}


// Get parameter, overwriting if already defined - return 1 if not found
int LTFitIOHandler::getParam(string key, string *valp)  {
  map<string,string>::iterator p;
  p = rpar.find(key);
  if ( p != rpar.end() )  {
    *valp = (*p).second;
    return 0;
  }
  else  {
    return 1;
  }
}

// Set parameter, overwriting if already defined
void LTFitIOHandler::setParam(string key, string val)  {
  map<string,string>::iterator p;
  p = rpar.find(key);
  if ( p != rpar.end() )  {  // change
    (*p).second = val;
  }
  else  {
    rpar[key] = val;
  }
}

void LTFitIOHandler::dbgShowParams(void)  {
  map<string, string>::iterator p;
  for(p=rpar.begin();p != rpar.end(); p++)  {
    cout << "PARAMETER:" << p->first << "=" << p->second << endl;
  }

  vector<string>::iterator q;
  for(q=snplist.begin();q != snplist.end();q++)  {
    cout << "SNP:" << *q << endl;
  }

}

void LTFitIOHandler::dbgShowCovariateData(void)  {
  
  Covariate *cp;
  DataPoint *dp;

  for(int i=0;i<ncovar;i++)  {

    cp = covariates[i];
    int nk = cp->nindices;
    int nj = cp->ndatapts;

    cout << "COVAR i = " << i << " NDATAPOINTS = " << nj << " NINDICES = " << nk << " TBAR = { ";
    for(int k=0;k<nk;k++)  {
      cout << cp->tbar[k] << " ";
    }
    cout << "}" << endl;

    for(int j=0;j<nj;j++)  {
      dp = cp->data[j];
      cout << "  point  { ";
      for(int k=0;k<nk;k++)  {
        if ( dp->missing[k] == 0 )
          cout << dp->T[k] << " ";
        else
          cout << "- ";
      }
      cout << "} liability = " << dp->sigma << endl;
    }

  }

}

void LTFitIOHandler::dbgShowIndivs(void)  {

  cout << "PHENOTYPES (nindiv = " << nindiv << ") = [";
  for(int i=0;i<nindiv;i++)  {
    cout << phenotypes[i] << " ";
  }
  cout << "]" << endl;

}





