#include "LTPrev.h"

int LTPrev::unpack(IOHandler *ioh)  {
  
  vector<Indiv *> V = ioh->getIndivData();
  ninds = V.size();

  // allocate memory
  pheno = new int[ninds];

  // get phenotypes
  vector<Indiv *>::iterator p;
  int i = 0;
  for(p=V.begin();p != V.end();p++)  {
    Indiv *ip = *p;
    pheno[i] = ip->pheno;
    i++;
  }

  string pstr; 
  if ( !ioh->getParam("prevalence",&pstr) )  {
    prevalence = atof(pstr.c_str());
  }
  else  {
    cerr << "error: no prevalence in parameter file" << endl;
    return 1;
  }

  return 0;

}

int LTPrev::run(IOHandler *ioh)  {

  // open genotype file for input
  string gfilename;
  if ( ioh->getParam("genotypefile",&gfilename) )  {
    cerr << "error: no genotypefile in parameter file" << endl;
    return 1;
  }

  ifstream genoinstream;
  genoinstream.open(gfilename.c_str(),ios::in);
  if ( !genoinstream.is_open() )  {
    cerr << "error: can't open genotypefile " << gfilename << endl;
    return 1;
  }

  // open output file
  string ofilename;
  if ( ioh->getParam("outfile",&ofilename) )  {
    cerr << "error: no output file in parameter file" << endl;
    return 1;
  }
  ofstream outstream;
  outstream.open(ofilename.c_str(),ios::out);
  if ( !outstream.is_open() )  {
    cerr << "error: can't open score output file " << ofilename << endl;
    return 1;
  }

  if ( inverse(prevalence, &mthreshold) )  {
    cerr << "error: can't find inverse of prevalence" << endl;
    return 1;
  }

  outstream << fixed << setw(10) << setprecision(6) << mthreshold << endl;  
  mthreshold = fabs(mthreshold);

  int *G = new int[ninds];

  while ( !genoinstream.eof() )  {

    if ( readGenotype(G, ninds, &genoinstream) ) continue;

    double sumcase = 0.0, sumctrl = 0.0;
    int countcase = 0, countctrl = 0;
    for(int i=0;i<ninds;i++)  {
      if ( G[i] == 9 ) 
        continue;
      if ( pheno[i] == 1 )  {       // case
        sumcase += (double) G[i];
        countcase += 1;
      }
      else if ( pheno[i] == 0 )  {  // ctrl
        sumctrl += (double) G[i];
        countctrl += 1;
      }
    }

    fcase = (countcase > 0 ? sumcase/countcase : 0.0);
    fctrl = (countctrl > 0 ? sumctrl/countctrl : 0.0);
    fexp = prevalence*fcase + (1.0-prevalence)*fctrl;

    int istat = 0;
    double epsilon = minimize(&istat);
   
    outstream << fixed << setw(10) << setprecision(6) 
              << fexp << "  " << epsilon << endl;  

  }

  outstream.close();
  delete G;
  return 0;

}


double LTPrev::f(double x)  {

  double p = fexp/2;
  double q = 1.0-p;
  double sigma = sqrt(1-2*p*q*x*x);

  double mu = x*(1-2*p);
  double pctrl1 = pnorm(mthreshold, mu, sigma, 1);

  mu = x*(2-2*p);
  double pctrl2 = pnorm(mthreshold, mu, sigma, 1);

  double pcase1 = 1.0 - pctrl1;
  double pcase2 = 1.0 - pctrl2;

  pctrl1 *= ( 2*p*q/(1.0-prevalence) );
  pctrl2 *= ( p*p/(1.0-prevalence) );

  pcase1 *= ( 2*p*q/prevalence );
  pcase2 *= ( p*p/prevalence );

  double result = (pctrl1 + 2*pctrl2 - fctrl)*(pctrl1 + 2*pctrl2 - fctrl) + 
                  (pcase1 + 2*pcase2 - fcase)*(pcase1 + 2*pcase2 - fcase);

  return (result); 

}

double LTPrev::minimize(int *ierr)  {

  /*
  for(int i=0;i<=1000;i++)  {
    double x = -1.0 + i*0.002;
    double fx = f(x);
  }
  return 0.0;
  */

  // Telescoping Grid search
  int N = 5;
  double xl = -1.0, xr = 1.0, dx = (xr - xl)/N; 
  do  {

    double fr = f(xr);
    double fl = f(xl);
    double fmin = ( fr < fl ? fr : fl);
    int imin = 0;

    for(int i=1;i<=N;i++)  {

      double x = xl + i*dx;
      double fx = f(x);
      if ( fx < fmin )  {
        imin = i;
        fmin = fx;
      }
    }

    if ( imin == 0 )  {
      xr = xl+dx;
    }
    else if ( imin == N )  {
      xl = xr-dx;
    }
    else  {
      xl = xl + (imin-1)*dx;
      xr = xl + (imin+1)*dx;
    }
    dx = (xr - xl)/N;

  } while ( dx > 1e-6 ); 
  return xl+0.5*dx;

}



// This is R function qnorm(x,0,1)
#define BIG 100.0
#define TINY 1.0e-10 
#define MAXITER 1000
int LTPrev::inverse(double y, double *result)  {
  double xl  = -BIG;
  double xr = +BIG;
  double fl = dnorm(xl) - y;
  double fr = dnorm(xr) - y;
  int iter = 0;

  while ( fl*fr > 0.0 )  {      // if HUGE isn't large enough, expand the search interval
    xl *= 2.0;
    xr *= 2.0;
    fl = dnorm(xl) - y;
    fr = dnorm(xr) - y;
    if ( iter++ > MAXITER )  {
      cerr << "error in binary search - can't find search interval"
           << "y = " << y << endl;
      return 1;
    }
  }

  double xmid = 0.5*(xl+xr);
  double fmid = dnorm(xmid) - y;
  iter = 0;
  while ( fabs(fmid) > TINY )  {
    if ( fl*fmid < 0 )  {
      xr = xmid;
      fr = fmid;
    }
    else if ( fmid*fr < 0 )  {
      xl = xmid;
      fl = fmid;
    }
    xmid = 0.5*(xl+xr);
    fmid = dnorm(xmid) - y;
    if ( iter++ > MAXITER )  {
      cerr << "error in binary search - can't find root" << endl;
      return 1;
    }
  }
  *result = xmid;
  return 0;

}

// This is really R function pnorm = integral from -Inf of Gaussian(0,1)
#define ROOT2 (sqrt(2.0))
double LTPrev::dnorm(double x)  {
  double y;
  if ( x > 0 )  {
    y = 1.0 + erf(x/ROOT2);
  }
  else  {
    y = 1.0 - erf(-x/ROOT2);
  }
  y *= 0.5;
  return(y);
}

// R function pnorm 
double LTPrev::pnorm(double x, double mu, double sigma, int lower_tail)  {
  double z = (x - mu)/sigma;
  double y = 0.5*(1.0 + erf(z/ROOT2));

  if ( lower_tail )  {
    return y;
  }
  else  {
    return 1.0-y;
  }
}

// This is duplicated in LTScore
int LTPrev::readGenotype(int *g, int n, ifstream *gin)  {
  if ( !gin->is_open() )  {
    cerr << "error: genotype file not open" << endl;
    return 1;
  }

  string line;
  getline(*gin,line);

  if ( gin->eof() ) {
    return -1;
  }

  if ( (int) line.length() < n)  {
    cerr << "error: line too short" << endl;
    return 1;
  }

  if ( (int) line.length() > n)  {
    cerr << "(warning) line too long length = " << line.length() << " n = " << n << endl;
  }

  for(int i=0;i<n;i++)  {
    int k = (int) line[i];
    k -= 48;                     // 48 is ASCII('0')
    if ( 0 <= k && k <= 2 )  {
      g[i] = k;
    }
    else  {
      g[i] = 9;
    }
  }
  return 0;
}

