#include "LTPub.h"



int LTPub::readInput(char *filename)  {
 
  ifstream fin;
  fin.open(filename,ios::in);
  if ( !fin.is_open() )  {
    cerr << "Can't open input file " << filename << endl;
    return 1;
  }

  // temporary vectors until we know how large to allocate arrays
  vector<Covariate *> cvtmp;
  vector<DataPoint *> dptmp;

  ncovar = 0;
  while ( !fin.eof() )  {

    Covariate *cp;
    DataPoint *dp;

    string line, origline, token;
    getline(fin,line);                      // get next line
    if ( line.empty() || fin.eof()  )  { 
      if ( !dptmp.empty() )  {               // put data points into previous covariate set
        int ndp = dptmp.size();
        cp->data = new DataPoint *[ndp];
        cp->ndatapts = ndp;
        for(int i=0;i<ndp;i++)  {
           cp->data[i] = dptmp[i];
        }
        dptmp.clear();
      }
      break; 
    }

    StringTokenizer stt(line," \t");
    stt.nextToken(token);                   //  get first token on line

    if ( !token.compare("MEAN") )  {        //  begin new covariate section
      if ( !dptmp.empty() )  {               // put data points into previous covariate set
        int ndp = dptmp.size();
        cp->data = new DataPoint *[ndp];
        cp->ndatapts = ndp;
        for(int i=0;i<ndp;i++)  {
           cp->data[i] = dptmp[i];
        }
        dptmp.clear();
      }

      ncovar++;
      int nind = stt.countTokens()-1;
      cp = new Covariate(nind);
      cvtmp.push_back(cp);

      for(int i=0;i<nind;i++)  {            // read means
        stt.nextToken(token); 
        cp->tbar[i] = atof(token.c_str());
        cp->C[i] = 0.0;
      }

    }
    else  {                                       // read data point into current covariate section
      if ( stt.countTokens() != (unsigned) (cp->nindices)+1 )  {       // should be nind covariates plus one prevalence
        cerr << "Error in data input" << endl;
        return 1;
      }

      dp = new DataPoint(cp->nindices);
      dptmp.push_back(dp);
      for(int i=0;i<cp->nindices;i++)  {    // read covariates
        if ( i > 0 )                        // first token on line already read
          stt.nextToken(token); 
        if ( !token.compare("NA") )  {
          dp->missing[i] = 1;
          dp->T[i] = 0.0;
        }
        else  {
          dp->missing[i] = 0;
          dp->T[i] = atof(token.c_str());
        }
      }
      stt.nextToken(token);          // read prevalence
      dp->sigma = atof(token.c_str());       // store it in sigma field for now
    }

  }

  covariates = new Covariate *[ncovar];
  for(int i=0;i<ncovar;i++)  {
    covariates[i] = cvtmp[i];
  }

  fin.close();
  return 0;

}




int LTPub::run(void)  {
  Covariate *cp;
  DataPoint *dp;
  double y;

  // compute sigmas from prevalence
  for(int i=0;i<ncovar;i++)  {
    cp = covariates[i];
    for(int k=0;k<cp->ndatapts;k++)  {
      dp = cp->data[k];
      if ( inverseDnorm(dp->sigma, &y) )  {
        cerr << "error: can't computing liability" << endl;
        return 1;
      }
      dp->sigma = y;
    }
  }
    
  // dbgShowInput();

  leastSquaresSolution(); 

  return 0;
}

void LTPub::leastSquaresSolution(void)  {

  int L = 1;
  for(int i=0;i<ncovar;i++)  {
    L += covariates[i]->nindices;
  }

  double *A = new double[L*L];
  double *b = new double[L];
  double *v = new double[L];

  // initialize 
  for(int i=0;i<L;i++)  {
    b[i] = 0.0;
    v[i] = 0.0;
    for(int j=0;j<L;j++)  {
      A[i*L+j] = 0.0;
    }
  }

  // upper-left elements
  int l = 0;
  for(int i1=0;i1<ncovar;i1++)  {
    for(int k1=0;k1<covariates[i1]->nindices;k1++)  {

      int m = 0;
      for(int i2=0;i2<ncovar;i2++)  {
        for(int k2=0;k2<covariates[i2]->nindices;k2++)  {
          if ( i1 == i2 )   {
  
            double a = 0.0;
            for(int j=0;j<covariates[i1]->ndatapts;j++)  {
              double z = (covariates[i1]->data[j]->T[k1] - covariates[i1]->tbar[k1]);
              double y = (covariates[i2]->data[j]->T[k2] - covariates[i2]->tbar[k2]);
              a += y*z;
            }

            A[l*L+m] = a;
          }
          m++;

        }
      }

      l++;
    }
  }

  // first L-1 rows, last column (coefficient of m in equation for C_ik) and
  // last row, first L-1 columns, last column (coefficients of C_ik in equation for m) and 
  // first L-1 rows of right-side vector (constant in equation or C_ik)

  l = 0;
  for(int i=0;i<ncovar;i++)  {
    for(int k=0;k<covariates[i]->nindices;k++)  {
      double a = 0.0;
      double e = 0.0;
      for(int j=0;j<covariates[i]->ndatapts;j++)  {
        double z = (covariates[i]->data[j]->T[k] - covariates[i]->tbar[k]);
        a += z;
        e += z*covariates[i]->data[j]->sigma; 
      }
      A[l*L+(L-1)] = a;     // note: some of the signs have been moved around
      A[(L-1)*L+l] = a;     // to make LTSOFT-0.6 reproduce LTSOFT-0.5 results
      b[l] = -e;    
      l++;
    }
  }

  // lower-left element in matrix (coefficient of m in equation for m)
  int ia = 0;
  for(int i=0;i<ncovar;i++)  {
    ia += covariates[i]->ndatapts;
  }
  A[(L-1)*L+(L-1)] = (double) ia;

  // last element in right-side vector
  double a = 0.0;
  for(int i=0;i<ncovar;i++)  {
    for(int j=0;j<covariates[i]->ndatapts;j++)  {
      a += covariates[i]->data[j]->sigma;
    }
  }
  b[L-1] = -a;

  LinearSolve(A,b,v,L);

  l = 0;
  for(int i=0;i<ncovar;i++)  {
    for(int k=0;k<covariates[i]->nindices;k++)  {
      a += covariates[i]->C[k] = -v[l];
      l++;
    }
  }
  mval = v[L-1];

  free(A);
  free(b);
  free(v);

}

void LTPub::writeOutput(char *filename)  {

  ofstream outstream;
  outstream.open(filename,ios::out);

  outstream << fixed << setw(10) << setprecision(4) << -mval << endl;
  for(int i=0;i<ncovar;i++)  {
    for(int k=0;k<covariates[i]->nindices;k++)  {
      outstream << fixed << setw(10) << setprecision(6) << covariates[i]->tbar[k] << "  ";
      outstream << fixed << setw(10) << setprecision(6) << covariates[i]->C[k] << "  ";
    }
    outstream << endl;
  }
  outstream.close();

}


// binary search for inverse of R function dnorm
#define BIG 100.0
#define TINY 1.0e-10 
#define MAXITER 1000
int LTPub::inverseDnorm(double y, double *result)  {
  double xl  = -BIG;
  double xr = +BIG;
  double fl = dnorm(xl) - y;
  double fr = dnorm(xr) - y;
  int iter = 0;

  while ( fl*fr > 0.0 )  {      // if HUGE isn't large enough, expand the search interval
    xl *= 2.0;
    xr *= 2.0;
    fl = dnorm(xl) - y;
    fr = dnorm(xr) - y;
    if ( iter++ > MAXITER )  {
      cerr << "error in binary search - can't find search interval" 
           << "y = " << y << endl;
      return 1;
    }
  }

  double xmid = 0.5*(xl+xr);
  double fmid = dnorm(xmid) - y;
  iter = 0;
  while ( fabs(fmid) > TINY )  {
    if ( fl*fmid < 0 )  {
      xr = xmid;
      fr = fmid;
    }
    else if ( fmid*fr < 0 )  {
      xl = xmid;
      fl = fmid;
    }
    xmid = 0.5*(xl+xr);
    fmid = dnorm(xmid) - y;
    if ( iter++ > MAXITER )  {
      cerr << "error in binary search - can't find root" << endl;
      return 1;
    }
  }
  *result = xmid;
  return 0;

}

double LTPub::evalF(void)  {

  double f = 0.0;
  for(int i=0;i<ncovar;i++)  {
    for(int j=0;j<covariates[i]->ndatapts;j++)  {

      double sigma_ij = covariates[i]->data[j]->sigma;
      double z = sigma_ij;
      for(int k=0;k<covariates[i]->nindices;k++)  {

        double tbar_ik = covariates[i]->tbar[k];
        double c_ik = covariates[i]->C[k];
        double t_ijk = covariates[i]->data[j]->T[k];
        z -= c_ik*(t_ijk - tbar_ik);

      }
      z += mval;
      f  += (z*z);

    }
  }
  return f; 

}


double LTPub::evalF(double *cval)  {

  double f = 0.0;

  // put array values into structs
  int l = 0;
  for(int i=0;i<ncovar;i++)  {
    for(int k=0;k<covariates[i]->nindices;k++)  {
      covariates[i]->C[k] = cval[l++];
    }
  }
  mval = cval[l];

  for(int i=0;i<ncovar;i++)  {
    for(int j=0;j<covariates[i]->ndatapts;j++)  {

      double sigma_ij = covariates[i]->data[j]->sigma;
      double z = sigma_ij;
      for(int k=0;k<covariates[i]->nindices;k++)  {

        double tbar_ik = covariates[i]->tbar[k];
        double c_ik = covariates[i]->C[k];
        double t_ijk = covariates[i]->data[j]->T[k];
        z -= c_ik*(t_ijk - tbar_ik);

      }
      z += mval;
      f  += (z*z);

    }
  }

  return f; 

}


// R function dnorm = integral from -Inf of Gaussian(0,1)
#define ROOT2 (sqrt(2.0))
double LTPub::dnorm(double x)  {
  double y;
  if ( x > 0 )  {
    y = 1.0 + erf(x/ROOT2);
  }
  else  {
    y = 1.0 - erf(-x/ROOT2);
  }
  y *= 0.5;
  return(y);
}

// --- DEBUGGING ---

void LTPub::dbgShowInput(void)  {
  Covariate *cp;
  DataPoint *dp;
  
  for(int i=0;i<ncovar;i++)  {

    cp = covariates[i];
    int nk = cp->nindices;
    int nj = cp->ndatapts;

    cout << "COVAR i = " << i << " NDATAPOINTS = " << nj << " NINDICES = " << nk << " TBAR = { ";
    for(int k=0;k<nk;k++)  {
      cout << cp->tbar[k] << " ";
    }
    cout << "}" << endl;

    for(int j=0;j<nj;j++)  {
      dp = cp->data[j];
      cout << "  point  { "; 
      for(int k=0;k<nk;k++)  {
        if ( dp->missing[k] == 0 )  
          cout << dp->T[k] << " ";
        else
          cout << "- ";
      }
      cout << "} liability = " << dp->sigma << endl;
    }

  }
}


#define KMAX 20 
#define DEBUG_SRCH 0
void LTPub::dbgGridSearch(void)  {

  ncval = 0;
  for(int i=0;i<ncovar;i++)  {
    ncval += covariates[i]->nindices;
  }
  ncval++;

  double *cval = new double[ncval];
  double *cmin = new double[ncval];
  double *cbak = new double[ncval];

  // save coefficient values because search will overwrite them
  int l = 0;
  for(int i=0;i<ncovar;i++)  {
    for(int k=0;k<covariates[i]->nindices;k++)  {
      cbak[l++] = covariates[i]->C[k];
    }
  }
  cbak[l] = mval;

  double fmin = evalF(cbak);
  cout << "    Least-Squares Solution: fmin = " << fixed << setprecision(4) << fmin << " C = [ ";
  for(int i=0;i<ncval-1;i++)  {
    cout << fixed << setprecision(4) << cbak[i] << "  ";
  }
  cout << "]  m = " << cbak[ncval-1] << endl;

  // exhaustive search
  fmin = HUGE_VAL;
  double *maxsrch = new double[ncval];
  double *minsrch = new double[ncval];
  double *delsrch = new double[ncval];

  for(int i=0;i<ncval;i++)  {
    minsrch[i] = -100.0;
    maxsrch[i] =  100.0;
    delsrch[i] =  5.0;
  }

  // for(int i=0;i<imax;i++)  {
  //  cmin[i] = cval[i];
  //}

  for(int k=0;k<=KMAX;k++)  {

    exhaustiveGridSearch(&fmin,cval,cmin,minsrch,maxsrch,delsrch,0);

    if ( DEBUG_SRCH || k == KMAX )  {
      cout << "Exhaustive Search Solution: fmin = " << fmin << " C = [ ";
      for(int i=0;i<ncval-1;i++)  {
        cout << cmin[i] << "  ";
      }
      cout << "]  m = " << cmin[ncval-1] << endl;
    }

    // telescope 
    for(int i=0;i<ncval;i++)  {
      minsrch[i] = cmin[i] - 2*delsrch[i];
      maxsrch[i] = cmin[i] + 2*delsrch[i];
      //  delsrch[i] /= 10.0;
      delsrch[i] /= 2.0;
    }

  }

  // replace coefficients
  l = 0;
  for(int i=0;i<ncovar;i++)  {
    for(int k=0;k<covariates[i]->nindices;k++)  {
      covariates[i]->C[k] = cbak[l++];
    }
  }
  mval = cbak[l];

  // clean up
  delete minsrch;
  delete maxsrch;
  delete delsrch;
  delete cval;
  delete cbak;
  delete cmin;

}


// recursive exhaustive search
void LTPub::exhaustiveGridSearch(double *fminp, double *cval, double *cmin, double *minsrch, double *maxsrch, double *delsrch, int n)  {

  double y, ymin = HUGE_VAL;
  double u, umin;

  double mn = minsrch[n];
  double mx = maxsrch[n];
  double d = delsrch[n];

  if ( n == ncval-1 )  {

    for(u=mn;u<=mx;u+=d)  {    // exhaustiveGridSearch over m

       cval[ncval-1] = u;
       y = evalF(cval);


       if ( DEBUG_SRCH && fabs(y) < 1.0 )  {
         cout << "f =  " << y << " C = [ ";
         for(int i=0;i<ncval-1;i++)  {
           cout << cval[i] << " ";
         }
         cout << "] m = " << cval[ncval-1] << endl;
       }


       if ( y < ymin )  {
         ymin = y;
         umin = u;
       }
    }

    if ( ymin < *fminp )  {        // if this is minimum, save the solution {c,m}

      if ( DEBUG_SRCH && fabs(y) < 1.0 ) {  cout << "  minimum = " << ymin << endl;  }
      *fminp = ymin;
      for(int i=0;i<ncval-1;i++)  {
         cmin[i] = cval[i];
      }
      cmin[ncval-1] = umin;

    }

  }
  else  {
    for(u=mn;u<=mx;u+=d)  {    // exhaustiveGridSearch over c[n]
      cval[n] = u;
      exhaustiveGridSearch(fminp,cval,cmin,minsrch,maxsrch,delsrch,n+1);
    }
  }
  return;

}

