#include "LTScore.h"
#include "LinearSolve.h"
#include "StringTokenizer.h"

#include <cmath>
#include <iomanip>
#include <vector>
 
#define DEBUG 0

int LTScore::unpackData(IOHandler *ioh)  {

  vector<Indiv *> V = ioh->getIndivData();
  vector<string> snplist = ioh->getSnplist();
  ncovar = ioh->getNcovars();
  neigs = ioh->getNeigs();
  nsamples = V.size();
  int n = ncovar + snplist.size();

  // allocate memory for clinical and genetic covariates
  pheno = new int[nsamples];
  if ( n > 0 )  {
    covariates = new double *[nsamples];
    missing = new bool *[nsamples];
    for(int i=0;i<nsamples;i++)  {
      covariates[i] = new double[n];
      missing[i] = new bool[n];
    }
  }

  // allocate memory for eigenvectors
  if ( neigs > 0 )  {
    evecs = new double *[nsamples];
    for(int j=0;j<neigs;j++)  {
      evecs[j] = NULL;
    }
  }

  // unpack clinical covariate data into arrays
  vector<Indiv *>::iterator p;
  int i = 0;

  int ecount = 0;
  int enullcount = 0;

  for(p=V.begin();p != V.end();p++)  {
    Indiv *ip = *p;
    pheno[i] = ip->pheno;
    if ( ip->pheno != -1 )  {
      for(int j=0;j<ncovar;j++)  {
        covariates[i][j] = ip->covariates[j]; 
        missing[i][j] = ip->missing[j]; 
      }
      if ( ip->evecs != NULL )  {
        evecs[i] = new double[neigs];
        for(int j=0;j<neigs;j++)  {
          evecs[i][j] = ip->evecs[j]; 
        }
        ecount++;
      }
      else  {
        enullcount++;
      }
    }
    i++;
  }

  // mean-adjust phenotype
  double sum = 0.0;
  int count = 0;
  phi = new double[nsamples];
  for(int i=0;i<nsamples;i++)  {
    if ( pheno[i] == -1 )
      continue;
    sum += (double) pheno[i];
    count++;
  }
  for(int i=0;i<nsamples;i++)  {
    if ( pheno[i] == -1 )
      continue;
    phi[i] = (double) pheno[i] - (sum/count);
  }

  if ( n > 0 )  
    ioh->getCovarParams(&mval, &beta, &tbar);


  // load genetic covariates from .geno file
  ngcovar = snplist.size();
  if ( ngcovar == 0 )  {
    return 0;
  }

  // read SNP list
  string f;
  int *index;
  if ( !ioh->getParam("snpfile",&f) )  {
    index = ioh->findSNPs(f.c_str());
    if ( index == NULL )  {
      return 1;
    }
  }
  else  {
    cerr << "error: SNP file must be supplied for genetic covariates" << endl;
    return 1;
  }

  // sort indices (assume few)
  for(int i=1;i<ngcovar;i++)  {
    for(int j=0;j<i;j++)  {
      if ( index[j] > index[i] )  {
        int t = index[j];
        index[j] = index[i];
        index[i] = t;
      }
    }
  }
  // squeeze out "not found"
  i=0;
  while ( index[i] == -1 )  {
    for(int j=1;j<ngcovar;j++)  {
      index[j-1] = index[j];
    }
    ngcovar--;
  }

  gcovarindices = index;

  // open genotype file for input
  string gfilename;
  if ( ioh->getParam("genotypefile",&gfilename) )  {
    cerr << "error: no genotypefile in parameter file" << endl;
    return 1;
  }

  ifstream genoinstream;
  genoinstream.open(gfilename.c_str(),ios::in);
  if ( !genoinstream.is_open() )  {
    cerr << "error: can't open genotypefile " << gfilename << endl;
    return 1;
  }

  int snpnum = -1;
  int snpidx = 0;
  int *G = new int[nsamples];

  while ( !genoinstream.eof() )  {
    snpnum++;

    if ( snpidx >= ngcovar ) {                      // no more genetic covariates to process
      break;
    }

    if ( readGenotype(G, nsamples, &genoinstream) ) continue;

    if ( snpnum < index[snpidx] )  {     // this SNP is not a genetic covariate
      continue;
    }

    for(int i=0;i<nsamples;i++)  {
      covariates[i][ncovar+snpidx] = G[i];
      missing[i][ncovar+snpidx] = ( G[i] == 9 );
    }
    snpidx++;

  }

  // mean-adjust genetic covariates 
  for(int i=0;i<ngcovar;i++)  {
    double sum = 0.0;
    int count = 0;
    for(int j=0;j<nsamples;j++)  {
      if ( !missing[j][ncovar+i] )  {
        sum += covariates[j][ncovar+i];
        count++;
      }
    }
    if ( count > 0 )  {
      for(int j=0;j<nsamples;j++)  {
        if ( !missing[j][ncovar+i] )  {
          covariates[j][ncovar+i] -= sum/count;
        }
      }
    }
  }

  delete G;
  return 0;
}


// constructor 
LTScore::LTScore(void)  {
  pheno = NULL;
  covariates = NULL;
  gcovariates = NULL;
  gcovarindices = NULL;
  missing = NULL;
  evecs = NULL;
  beta = NULL;
  tbar = NULL;
  phi = NULL;
  phiadj = NULL;
  xi = NULL;
  xiadj = NULL;
  geno = NULL;
  gamma = NULL;
  gammaadj = NULL;
}

// destructor 
LTScore::~LTScore(void)  {
  if ( covariates != NULL )  {
    for(int i=0;i<nsamples;i++)  
      if ( covariates[i] != NULL ) 
        delete covariates[i]; 
  }
  if ( gcovariates != NULL )  {
    for(int i=0;i<nsamples;i++)  
      if ( gcovariates[i] != NULL ) 
        delete gcovariates[i]; 
  }
  if ( missing != NULL )  {
    for(int i=0;i<nsamples;i++)  
      if ( missing[i] != NULL ) 
        delete missing[i]; 
  }
  if ( evecs != NULL )  {
    for(int i=0;i<nsamples;i++)  
      if ( evecs[i] != NULL ) 
        delete evecs[i]; 
  }

  if ( gcovarindices != NULL )  delete gcovarindices;
  if ( pheno != NULL )  delete pheno;
  if ( phi != NULL ) delete phi;
  if ( phiadj != NULL ) delete phiadj;
  if ( xi != NULL ) delete xi;
  if ( xiadj != NULL ) delete xiadj;
  if ( geno != NULL )  delete geno;
  if ( gamma != NULL )  delete gamma;
  if ( gammaadj != NULL )  delete gammaadj;

  // do not free beta and tbar because they will be freed in 
  // IOHandler destructor

}

int LTScore::run(IOHandler *ioh)  {

  // open genotype file for input
  string gfilename;
  if ( ioh->getParam("genotypefile",&gfilename) )  {
    cerr << "error: no genotypefile in parameter file" << endl;
    return 1;
  }

  ifstream genoinstream;
  genoinstream.open(gfilename.c_str(),ios::in);
  if ( !genoinstream.is_open() )  {
    cerr << "error: can't open genotypefile " << gfilename << endl;
    return 1;
  }

  // open snpfile so we can print rsid to output
  string sfilename;
  if ( ioh->getParam("snpfile",&sfilename) )  {
    cerr << "error: no snpfile in parameter file" << endl;
    return 1;
  }

  ifstream snpinstream;
  snpinstream.open(sfilename.c_str(),ios::in);
  if ( !snpinstream.is_open() )  {
    cerr << "error: can't open snpfile " << sfilename << endl;
    return 1;
  }

  // open score output file
  string sofilename;
  if ( ioh->getParam("scoreoutfile",&sofilename) )  {
    cerr << "error: no score output file in parameter file" << endl;
    return 1;
  }
  ofstream scoreoutstream;
  scoreoutstream.open(sofilename.c_str(),ios::out);
  if ( !scoreoutstream.is_open() )  {
    cerr << "error: can't open score output file " << sofilename << endl;
    return 1;
  }

  // Things necessary to compute LTSCORE - skip if only EIGENSTRAT
  if ( ncovar > 0 )  {  

    // assign case/control mean values to missing covariates
    int istat = estimateMissingCovariates();
    if ( istat )
      return 1;
  }

  if ( ncovar+ngcovar > 0 )  {
    int istat;

    // compute posterior mean liability for each sample
    xi = new double[nsamples];      // posterior mean liability
    if ( ( istat = computeLiability(xi) ) )   {
      return 1;
    }

    // print posterior means to output file
    string pmfilename;
    if ( !ioh->getParam("posteriormeanfile",&pmfilename) )  {
      ofstream postmeanoutstream;
      postmeanoutstream.open(pmfilename.c_str(),ios::out);
      for(int k=0;k<nsamples;k++)  {
        postmeanoutstream << xi[k] << endl;
      }
      postmeanoutstream.close();
    }

  }

  if ( ncovar > 0 || neigs > 0 )  {
    adjustOLSvecs();
  }

  geno = new int[nsamples];            // genotype
  gamma = new double[nsamples];        // mean-adjusted genotype
  if ( neigs > 0 )  {
    gammaadj = new double[nsamples];
  }

  scoreoutstream << "  SNPID           ATT     ";
  if ( ncovar+ngcovar > 0 )
    scoreoutstream <<  "   LTSCORE  "; 
  if ( neigs > 0 )
    scoreoutstream <<  " EIGENSTRAT ";
  if ( ncovar+ngcovar > 0 && neigs > 0 )
    scoreoutstream <<  "   LT-EIG   ";
  if ( neigs > 0 )
    scoreoutstream <<  " OLS ";
  if ( ncovar+ngcovar > 0 && neigs > 0 )
    scoreoutstream <<  "   LT-OLS   ";
  scoreoutstream << endl;

  int snpindex = -1;
  while ( !genoinstream.eof() )  {

    // read RSID so it can be printed to output file
    string snpline, snpname;
    getline(snpinstream,snpline);
    StringTokenizer st(snpline," \t");
    if ( snpline.empty() || snpinstream.eof() || st.countTokens() == 0 )  {
      snpname = "unknown";
    }
    else  {
      st.nextToken(snpname);
    }

    snpindex++;
    if ( readGenotype(geno, nsamples, &genoinstream) ) continue;

    // if this SNP is a genetic covariate, print NA instead of score
    int skip = 0;
    for(int i=0;i<ngcovar;i++)  {
      if ( snpindex == gcovarindices[i] )  {
        scoreoutstream << fixed << setw(12) << snpname;
        scoreoutstream << "          NA"; 
        if ( ncovar+ngcovar > 0 )  scoreoutstream << "          NA"; 
        if ( neigs > 0 )   scoreoutstream << "          NA"; 
        if ( neigs > 0 && ncovar+ngcovar > 0 )  scoreoutstream << "          NA"; 
        if ( neigs > 0 )   scoreoutstream << "          NA";
        if ( neigs > 0 && ncovar+ngcovar > 0 )    scoreoutstream << "          NA";
        scoreoutstream << endl;
        skip = 1;
      }
    }
    if ( skip ) continue;

    meanAdjustGeno();
    if ( neigs > 0 )  {
      adjustOLSGeno();
    }
    double score1, score2, score3, score4, score5, score6;

    // question: why am I passing class members pheno and nsamples to these routines?
    if ( computeATTScore(geno,pheno,nsamples,&score1) )  continue;
    if ( ncovar+ngcovar > 0 ) 
      if ( computeLTScore(geno,pheno,xi,nsamples,&score2) )  continue;
    if ( neigs > 0 ) 
      if ( computeEIGScore(geno,pheno,nsamples,&score3) )  continue;
    if ( ncovar+ngcovar > 0 && neigs > 0 ) 
      if ( computeEIGLTScore(geno,pheno,xi,nsamples,&score4) )  continue;
    if ( neigs > 0 ) 
      if ( computeOLSScore(geno,pheno,nsamples,&score5) )  continue;
    if ( ncovar+ngcovar > 0 && neigs > 0 ) 
      if ( computeOLSLTScore(geno,pheno,xi,nsamples,&score6) )  continue;
    
    scoreoutstream << fixed << setw(12) << snpname;
    scoreoutstream << fixed << setw(12) << setprecision(4) << score1;
    if ( ncovar+ngcovar > 0 )  
      scoreoutstream << fixed << setw(12) << setprecision(4) << score2;
    if ( neigs > 0 )  
      scoreoutstream << fixed << setw(12) << setprecision(4) << score3;
    if ( neigs > 0 && ncovar+ngcovar > 0 )  
      scoreoutstream << fixed << setw(12) << setprecision(4) << score4;
    if ( neigs > 0 )  
      scoreoutstream << fixed << setw(12) << setprecision(4) << score5;
    if ( neigs > 0 && ncovar+ngcovar > 0 )  
      scoreoutstream << fixed << setw(12) << setprecision(4) << score6;
    scoreoutstream << endl;

  }

  snpinstream.close();
  scoreoutstream.close();
  return 0;

}


int LTScore::readGenotype(int *g, int n, ifstream *gin)  {
  if ( !gin->is_open() )  {
    cerr << "error: genotype file not open" << endl;
    return 1;
  } 
  
  string line;
  getline(*gin,line);
  
  if ( gin->eof() ) {
    return -1;
  }

  if ( (int) line.length() < n)  {
    cerr << "error: line too short" << endl;
    return 1;
  }

  if ( (int) line.length() > n)  {
    cerr << "(warning) line too long length = " << line.length() << " n = " << n << endl;
  }

  for(int i=0;i<n;i++)  {
    int k = (int) line[i];
    k -= 48;                     // 48 is ASCII('0')
    if ( 0 <= k && k <= 2 )  {
      g[i] = k;
    }
    else  {
      g[i] = 9;
    }
  }
  return 0;
}



int LTScore::estimateMissingCovariates(void)  {

  for(int i=0;i<ncovar;i++)  {

    double sumCase = 0.0;
    double sumCtrl = 0.0;
    int nCasePresent = 0;
    int nCaseMissing = 0;
    int nCtrlPresent = 0;
    int nCtrlMissing = 0;

    // compute mean in Cases or Controls
    for(int j=0;j<nsamples;j++)  {
      if ( pheno[j] != 0 && pheno[j] != 1 ) continue;
      if ( pheno[j] == 0 )  {
        if ( missing[j][i] )   {
          nCtrlMissing++;
        } 
        else  {
          sumCtrl += covariates[j][i];
          nCtrlPresent++;
        }
      }
      else if ( pheno[j] == 1 )  {
        if ( missing[j][i] )  {
          nCaseMissing++;
        }
        else  {
          sumCase += covariates[j][i];
          nCasePresent++;
        }
      }
    }

    if ( nCaseMissing > 0 && nCasePresent == 0 )  {
      cout << "error: no cases to estimate mean of covariate " << i << " from" << endl;
      return 1;
    }

    if ( nCtrlMissing > 0 && nCtrlPresent == 0 )  {
      cout << "error: no controls to estimate mean of covariate " << i << " from" << endl;
      return 1;
    }

    // assign mean in Cases or Controls respectively to missing Case or Control covariate
    for(int j=0;j<nsamples;j++)  {
      if ( pheno[j] != 0 && pheno[j] != 1 ) continue;
      if ( missing[j][i] )  {
        covariates[j][i] = ( pheno[j] ? sumCase/nCasePresent : sumCtrl/nCtrlPresent );
      }
    }

  }

  return 0;

}


#define ROOT2 (sqrt(2.0))
#define RECIPROOT2PI (1.0/sqrt(2.0*3.14159265))
int LTScore::computeLiability(double *x)  {

  for(int i=0;i<nsamples;i++)  {
    x[i] = 0.0;
    if ( pheno[i] != 0 && pheno[i] != 1 ) continue;
    for(int j=0;j<ncovar+ngcovar;j++)  {
      x[i] += beta[j]*( covariates[i][j] - tbar[j] );
    }
    x[i] += mval;
  }

  double numer, denom;
  for(int i=0;i<nsamples;i++)  {
    double y = -x[i];
    if ( pheno[i] == 1 )  {   // Case
      numer = RECIPROOT2PI*exp(-y*y/2.0);
      denom = ( y > 0 ? 1.0 - erf(y/ROOT2) : 1.0 + erf(-y/ROOT2) );
    }
    else if ( pheno[i] == 0 )  {  // Control
      numer = -RECIPROOT2PI*exp(-y*y/2.0);
      denom = ( y > 0 ? 1.0 + erf(y/ROOT2) : 1.0 - erf(-y/ROOT2) );
    }
    x[i] = numer/(denom/2.0);
  }

  return 0;

}


inline int LTScore::computeATTScore(int *G, int *P, int n, double *result)  {

  double sumx = 0.0, sumy = 0.0, sumxx = 0.0, sumxy = 0.0, sumyy = 0.0;
  int  sum1 = 0;

  // do this here, because these are int-type vectors
  for(int i=0;i<n;i++)  {
    if ( pheno[i] <= -1 || geno[i] == 9 ) {
      continue;
    }
    sumx  += geno[i];
    sumy  += pheno[i];
    sumxx += (geno[i]*geno[i]);
    sumxy += (geno[i]*pheno[i]);
    sumyy += (pheno[i]*pheno[i]);
    sum1  += 1;
  }

  if ( sum1 == 0 )  {
    cout << "error: no data" << endl;
    return 1;
  }
  
  double mx = sumx/sum1;
  double my = sumy/sum1;
  double sx = sqrt(sumxx/sum1-mx*mx);
  double sy = sqrt(sumyy/sum1-my*my);
  if ( fabs(sx) < 1e-8 || fabs(sy) < 1e-8 )  {
    cout << "error: zero variance" << endl;
    return 1;
  }

  double rho = (sumxy/sum1 - mx*my)/(sx*sy);
  *result = sum1*rho*rho;
  return 0;

}

inline int LTScore::computeLTScore(int *G, int *P, double *X, int n, double *result)  {

  double sumx = 0.0, sumy = 0.0, sumxx = 0.0, sumxy = 0.0, sumyy = 0.0;
  int  sum1 = 0;

  // do this here, because geno is int-type vector
  for(int i=0;i<n;i++)  {
    if ( pheno[i] <= -1 || geno[i] == 9 ) {
      continue;
    }
    sumx  += geno[i];
    sumy  += xi[i];
    sumxx += (geno[i]*geno[i]);
    sumxy += (geno[i]*xi[i]);
    sumyy += (xi[i]*xi[i]);
    sum1  += 1;
  }

  if ( sum1 == 0 )  {
    cout << "error: no data" << endl;
    return 1;
  }

  double mx = sumx/sum1;
  double my = sumy/sum1;
  double sx = sqrt(sumxx/sum1-mx*mx);
  double sy = sqrt(sumyy/sum1-my*my);
  if ( fabs(sx) < 1e-8 || fabs(sy) < 1e-8 )  {
    cout << "error: zero variance" << endl;
    return 1;
  }
  double rho = (sumxy/sum1 - mx*my)/(sx*sy);
  *result = sum1*rho*rho;
  return 0;

}

inline int LTScore::computeEIGScore(int *G, int *P, int n, double *result)  {

  double *ephi = new double[n];      // PC-adjusted phenotype
  double *egamma = new double[n];    // PC-adjusted phenotype

  for(int i=0;i<nsamples;i++)  {
    egamma[i] = gamma[i];
    ephi[i] = phi[i];
  }

  // adjust phi for PC's
  for(int k=0;k<neigs;k++)  {
    double z = 0.0, y = 0.0;

    for(int i=0;i<n;i++)  {
      if ( pheno[i] == -1 || evecs[i] == NULL ) 
        continue;
      z += evecs[i][k]*ephi[i];
      y += evecs[i][k]*evecs[i][k];
    }
    double ap = ( y > 0 ? (z*y)/(y*y) : 0.0 );
    for(int i=0;i<n;i++)  {
      if ( pheno[i] == -1 || evecs[i] == NULL ) 
        continue;
      ephi[i] -= ap*evecs[i][k];
    }
  }
  // adjust gamma for PC's
  for(int k=0;k<neigs;k++)  {
    double z = 0.0, y = 0.0;

    for(int i=0;i<n;i++)  {
      if ( geno[i] == 9 || evecs[i] == NULL ) 
        continue;
      z += evecs[i][k]*egamma[i];
      y += evecs[i][k]*evecs[i][k];
    }
    double ap = ( y > 0 ? (z*y)/(y*y) : 0.0 );
    for(int i=0;i<n;i++)  {
      if ( geno[i] == 9 || evecs[i] == NULL ) 
        continue;
      egamma[i] -= ap*evecs[i][k];
    }
  }

  int sum1 = 0;
  for(int i=0;i<n;i++)  {
    if ( pheno[i] <= -1 || geno[i] == 9 || evecs[i] == NULL ) 
      continue;
    sum1 += 1;
  }

  double rho;
  int ierr = correlation(egamma, ephi, &rho);
  if ( ierr )  {
    cout << "error: zero variance" << endl;
    return 1;
  }
  double s = sum1 - (double)(neigs+1);
  *result = s*rho*rho;
  free(ephi);
  free(egamma);
  return 0;

}

inline int LTScore::computeEIGLTScore(int *G, int *P, double *X, int n, double *result)  {

  // eigenstrat : X => X-(V*X)/(V*V)  
  // P is used only as a missing-data flag
  // QUESTION: Do we need to mean-adjust X and G first?

  double *egamma = new double[n];  // PC-adjusted genotype
  double *exi = new double[n];    // PC-adjusted genotype

  for(int i=0;i<nsamples;i++)  {
    egamma[i] = gamma[i];
    exi[i] = xi[i];
  }

  // mean-adjust phenotype (posterior mean)
  double sum = 0.0;
  int count = 0;
  for(int i=0;i<n;i++)  {
    if ( P[i] == -1 )
      continue;
    sum += (double) xi[i];
    count++;
  }
  for(int i=0;i<n;i++)  {
    if ( P[i] == -1 )
      continue;
    exi[i] = (double) xi[i] - (sum/count);
  }

  // adjust xi for PC's
  for(int k=0;k<neigs;k++)  {
    double z = 0.0, y = 0.0;

    for(int i=0;i<n;i++)  {
      if ( pheno[i] == -1 || evecs[i] == NULL ) 
        continue;
      z += evecs[i][k]*exi[i];
      y += evecs[i][k]*evecs[i][k];
    }
    double ap = ( y > 0 ? (z*y)/(y*y) : 0.0 );
    for(int i=0;i<n;i++)  {
      if ( pheno[i] == -1 || evecs[i] == NULL ) 
        continue;
      exi[i] -= ap*evecs[i][k];
    }
  }

  // adjust gamma for PC's
  for(int k=0;k<neigs;k++)  {
    double z = 0.0, y = 0.0;

    for(int i=0;i<n;i++)  {
      if ( geno[i] == 9 || evecs[i] == NULL ) 
        continue;
      z += evecs[i][k]*egamma[i];
      y += evecs[i][k]*evecs[i][k];
    }
    double ap = ( y > 0 ? (z*y)/(y*y) : 0.0 );
    for(int i=0;i<n;i++)  {
      if ( geno[i] == 9 || evecs[i] == NULL ) 
        continue;
      egamma[i] -= ap*evecs[i][k];
    }
  }

  int sum1 = 0;
  for(int i=0;i<n;i++)  {
    if ( pheno[i] <= -1 || geno[i] == 9 || evecs[i] == NULL ) 
      continue;
    sum1 += 1;
  }
  double rho;
  int ierr = correlation(egamma, exi, &rho);
  if ( ierr )  {
    cout << "error: zero variance" << endl;
    return 1;
  }
  double s = sum1 - (double)(neigs+1);
  *result = s*rho*rho;

  free(egamma);
  free(exi);

  return 0;
}


inline int LTScore::computeOLSScore(int *G, int *P, int n, double *result)  {

  double rho;
  int ierr = correlation(gammaadj, phiadj, &rho);
  if ( ierr )  {
    cout << "error: zero variance" << endl;
    return 1;
  }
  int sum1 = 0;
  for(int i=0;i<n;i++)  {
    if ( pheno[i] <= -1 || geno[i] == 9 || evecs[i] == NULL ) 
      continue;
    sum1 += 1;
  }
  double s = sum1 - (double)(neigs+1);
  *result = s*rho*rho;
  return 0;

}

inline int LTScore::computeOLSLTScore(int *G, int *P, double *X, int n, double *result)  {
  double rho;

  int ierr = correlation(gammaadj, xiadj, &rho);
  if ( ierr )  {
    cout << "error: zero variance" << endl;
    return 1;
  }
  int sum1 = 0;
  for(int i=0;i<n;i++)  {
    if ( pheno[i] <= -1 || geno[i] == 9 || evecs[i] == NULL ) 
      continue;
    sum1 += 1;
  }
  double s = sum1 - (double)(neigs+1);
  *result = s*rho*rho;
  return 0;

}

int LTScore::adjustOLSvecs(void)  {

  if ( neigs == 0 )
    return 0;

  xtx = new double[neigs*neigs];
  double *q = new double[neigs];     // (XTX)*pheno for linear solve
  double *eta = new double[neigs];  

  for(int k=0;k<neigs;k++)  {
    for(int l=0;l<=k;l++)  {
      if ( evecs[l] == NULL || evecs[k] == NULL )
        continue;
      xtx[k*neigs+l] = 0.0;      
      for(int i=0;i<nsamples;i++)  {
        xtx[k*neigs+l] += (evecs[i][k] * evecs[i][l]); 
      }
      xtx[l*neigs+k] = xtx[k*neigs+l];
    }
  }

  for(int k=0;k<neigs;k++)  {
    q[k] = 0.0;      
    for(int i=0;i<nsamples;i++)  {
      if ( evecs[k] == NULL || pheno[i] == -1 )
        continue;
      // q[k] += (evecs[i][k] * pheno[i]);  
      q[k] += (evecs[i][k] * phi[i]);  
    }
  }
  LinearSolve::LinearSolve(xtx, q, eta, neigs);
  phiadj = new double[nsamples];
  for(int i=0;i<nsamples;i++)  {
    if ( pheno[i]  == -1 || evecs[i] == NULL )  
      continue;
    double x= 0.0;
    for(int j=0;j<neigs;j++)  {
      x += eta[j]*evecs[i][j];
    }
    // phiadj[i] = pheno[i] - x;
    phiadj[i] = phi[i] - x;
  }

  if ( ncovar > 0 )  {

    for(int k=0;k<neigs;k++)  {
      q[k] = 0.0;      
      for(int i=0;i<nsamples;i++)  {
        if ( evecs[k] == NULL || pheno[i] == -1 )
          continue;
        q[k] += (evecs[i][k] * xi[i]);  
      }
    }
    LinearSolve::LinearSolve(xtx, q, eta, neigs);
    xiadj = new double[nsamples];
    for(int i=0;i<nsamples;i++)  {
      if ( pheno[i]  == -1 || evecs[i] == NULL )  
        continue;
      double x= 0.0;
      for(int j=0;j<neigs;j++)  {
        x += eta[j]*evecs[i][j];
      }
      xiadj[i] = xi[i] - x;
    }

  }

  delete q;
  delete eta;
  return 0;

}

void LTScore::meanAdjustGeno(void)  {

  // mean-adjust genotype
  double sum = 0.0;
  int count = 0;
  for(int i=0;i<nsamples;i++)  {
    if ( geno[i] == 9 )
      continue;
    sum += (double) geno[i]/2.0;
    count += 1;
  }
  for(int i=0;i<nsamples;i++)  {
    if ( geno[i] == 9 )
      continue;
    gamma[i] = (double) geno[i]/2.0 - (sum/count);
  }
}

void LTScore::adjustOLSGeno(void)  {

  double *q = new double[neigs];     // (XTX)*pheno for linear solve
  double *eta = new double[neigs];  
  for(int k=0;k<neigs;k++)  {
    q[k] = 0.0;      
    for(int i=0;i<nsamples;i++)  {
      if ( evecs[k] == NULL || pheno[i] == -1 )
        continue;
      q[k] += (evecs[i][k] * gamma[i]);  
    }
  }
  LinearSolve::LinearSolve(xtx, q, eta, neigs);
  for(int i=0;i<nsamples;i++)  {
    if ( pheno[i]  == -1 || evecs[i] == NULL )  
      continue;
    double x= 0.0;
    for(int j=0;j<neigs;j++)  {
      x += eta[j]*evecs[i][j];
    }
    gammaadj[i] = gamma[i] - x;
  }
  delete q;
  delete eta;

}



int LTScore::correlation(double *X, double *Y, double *result)  {
  double sumx = 0.0, sumy = 0.0, sumxx = 0.0, sumxy = 0.0, sumyy = 0.0;
  int sum1 = 0;

  for(int i=0;i<nsamples;i++)  {

    if ( pheno[i] <= -1 || geno[i] == 9 || evecs[i] == NULL ) 
      continue;

    sumx += X[i];
    sumy += Y[i];
    sumxx += (X[i]*X[i]);
    sumxy += (X[i]*Y[i]);
    sumyy += (Y[i]*Y[i]);
    sum1++;
  }

  double mx = sumx/sum1;
  double my = sumy/sum1;
  double sx = sqrt(sumxx/sum1-mx*mx);
  double sy = sqrt(sumyy/sum1-my*my);

  if ( fabs(sx) < 1e-8 || fabs(sy) < 1e-8 )  {
    cout << "error: zero variance" << endl;
    return 1;
  }
  *result = (sumxy/sum1 - mx*my)/(sx*sy);
  return 0;

}
